
/*
   smtp_socket.h
   this is a simple interface to sending E-Mails from any C++ application.
   it only supports the (insecure) LOGIN authentification method so far.
   
   this sourcecode is provided under the terms of the GNU general public
   license version 3. See gpl3.txt for details.
   
   example of usage:
   =================
   SMTP s("server:port", "my_username", "my_password");
   EMail m("my_email_address", "recipients_email_address", "the subject");
   m += "hi there";
   m += "how are you?";
   s.Send(m);
   
   hardcore example:
   =================
   SMTP("server:port", "username", "password").Send(EMail("sender", "recipient", "subject", "text"));
      
   
   you need to give the compiler the -lpthread option, because the used socket
   class is able to work actively (running in its own thread), although that
   feature is not used in this code.
   
   thanks to
      http://www.faqs.org/rfcs/rfc821.html
      http://de.wikipedia.org/wiki/SMTP-Auth     (for authentification)
      http://www.hochfeiler.it/alvise/smtp.htm   (for attachments)
      http://www.ietf.org/rfc/rfc2821.txt        (for BCC)
*/

#include "smtp_socket.h"
using namespace std;

namespace AlgorithMan
{
    // -------------------------------------------------------------------------
    
    EMail::EMail(string Sender, string Recipient, string Subject, string Text)
    {
        this->Sender = Sender;
        this->Recipients = vector<string>(1,Recipient);
        this->Subject = Subject;
        this->Text = vector<string>(1,Text);
    }
    
    EMail::EMail(string Sender, string Recipient, string Subject, vector<string> Text)
    {
        this->Sender = Sender;
        this->Recipients = vector<string>(1,Recipient);
        this->Subject = Subject;
        this->Text = Text;
    }
    
    EMail::EMail(string Sender, vector<string> Recipients, string Subject, string Text)
    {
        this->Sender = Sender;
        this->Recipients = Recipients;
        this->Subject = Subject;
        this->Text = vector<string>(1,Text);
    }
    
    EMail::EMail(string Sender, vector<string> Recipients, string Subject, vector<string> Text)
    {
        this->Sender = Sender;
        this->Recipients = Recipients;
        this->Subject = Subject;
        this->Text = Text;
    }
    
    EMail::EMail(string Sender, string Recipient, string Subject)
    {
        this->Sender = Sender;
        this->Recipients = vector<string>(1,Recipient);
        this->Subject = Subject;
    }
    
    EMail::EMail(string Sender, vector<string> Recipients, string Subject)
    {
        this->Sender = Sender;
        this->Recipients = Recipients;
        this->Subject = Subject;
    }
    
    void EMail::operator+=(string line)
    {
        Text.push_back(line);
    }
    
    void EMail::AddRecipient(string Recipient)
    {
        Recipients.push_back(Recipient);
    }
    
    void EMail::AddBlindCopyRecipient(string BCCRecipient)
    {
        BCC.push_back(BCCRecipient);
    }
    
    void EMail::AddAttachment(string path_to_file)
    {
        attached_files.push_back(path_to_file);
    }
    
    // -------------------------------------------------------------------------

    SMTP::SMTP(string server, string username, string password)
    {
        this->server = server;
        this->username = username;
        this->password = password;
    }

    SMTP::~SMTP()
    {
    }

    void SMTP::Send(vector<EMail> Mails)
    {
        Connect(server);
        // handshake
        if(username == "")  // server does not require auth
        {
            Socket::Send("HELO foobar\r\n");
        }
        else
        {
            Socket::Send("EHLO foobar\r\n");
            Socket::Send("AUTH PLAIN\r\n");
            Socket::Send(string("{")+to_string(username.size()+password.size()+2)+"}\r\n");


            unsigned char *plainstr = new(unsigned char[username.size()+password.size()+2]);
            plainstr[0] = 0;
            memcpy(plainstr+1, username.c_str(), username.size()*sizeof(unsigned char));
            plainstr[username.size()+1] = '\0';
            memcpy(plainstr+username.size()+2, password.c_str(), password.size()*sizeof(unsigned char));
            Socket::Send(ToBase64(plainstr, username.size()+password.size()+2)+"\r\n");
            delete[](plainstr);

/*
            vector<unsigned char> plainstr(username.size()+password.size()+2, ' ');
            plainstr[0] = 0;
            for(int i = 0; i < username.size(); i++)
                plainstr[i+1] = username[i];
            plainstr[username.size()+1] = 0;
            for(int i = 0; i < password.size(); i++)
                plainstr[username.size()+2+i] = password[i];
            Socket::Send(ToBase64(&plainstr[0], username.size()+password.size()+2)+"\r\n");
*/
        }

        for(vector<EMail>::iterator i = Mails.begin(); i != Mails.end(); i++)
        {
            // this looks stupid, but it's the most secure BCC method to send
            // individual mails to all BCC recipients
            vector<vector<string> > individual_mails;
            individual_mails.push_back(i->Recipients);
            for(vector<string>::iterator bcc_rcpt = i->BCC.begin(); bcc_rcpt != i->BCC.end(); bcc_rcpt++)
            {
                individual_mails.push_back(vector<string>(1,*bcc_rcpt));
            }

            // iterate individual mails
            for(vector<vector<string> >::iterator rcpts = individual_mails.begin(); rcpts != individual_mails.end(); rcpts++)
            {
                // send message head
                Socket::Send(string("MAIL FROM:")+i->Sender+"\r\n");
                for(vector<string>::iterator rcpt = rcpts->begin(); rcpt != rcpts->end(); rcpt++)
                {
                    Socket::Send(string("RCPT TO:")+(*rcpt)+"\r\n");
                }
                Socket::Send("DATA\r\n");
                Socket::Send(string("From: <")+i->Sender+"> \r\n");
                for(vector<string>::iterator rcpt = i->Recipients.begin(); rcpt != i->Recipients.end(); rcpt++)
                    Socket::Send(string("To: <")+(*rcpt)+"> \r\n");
                if(i->Subject != "")
                    Socket::Send(string("Subject: ")+i->Subject+" \r\n");
                if(i->attached_files.size() > 0)
                {
                    Socket::Send("MIME-Version: 1.0 \r\n");
                    Socket::Send("Content-Type: multipart/mixed; \r\n");
                    // this "boundary" idea is so stupid! who came up with this shit!?
                    Socket::Send(" boundary= \"VGhlIFNNVFAgUHJvdG9jb2wgaXMgY3JhcCBhbmQgd2FzIGludmVudGVkIGJ5IGFuIGlkaW90\" \r\n");
                    Socket::Send("\r\n");
                    Socket::Send("--VGhlIFNNVFAgUHJvdG9jb2wgaXMgY3JhcCBhbmQgd2FzIGludmVudGVkIGJ5IGFuIGlkaW90 \r\n");
                    Socket::Send("Content-Type: text/plain; charset=US-ASCII \r\n");
                }

                // send message body
                Socket::Send("\r\n");
                for(vector<string>::iterator txt = i->Text.begin(); txt != i->Text.end(); txt++)
                {
                    Socket::Send((*txt)+"\r\n");
                }
                Socket::Send("\r\n");
                
                // send file attachments
                for(vector<string>::iterator afile = i->attached_files.begin(); afile != i->attached_files.end(); afile++)
                {
                    Socket::Send("--VGhlIFNNVFAgUHJvdG9jb2wgaXMgY3JhcCBhbmQgd2FzIGludmVudGVkIGJ5IGFuIGlkaW90 \r\n");
                    Socket::Send("Content-Type: application/octet-stream \r\n");
                    Socket::Send("Content-Transfer-Encoding: base64 \r\n");
                    Socket::Send("Content-Disposition: attachment; \r\n");
                    Socket::Send(string(" filename= \"")+basename(*afile)+"\" \r\n\r\n");

                    FILE* fin = fopen((*afile).c_str(), "rb");
                    unsigned char filebuffer[58];
                    for(int n = 57; n >= 57;)
                    {
                        n = fread(&filebuffer[0],sizeof(unsigned char),57,fin);
                        Socket::Send(ToBase64(&filebuffer[0], n)+"\r\n");
                    }
                    fclose(fin);
                }
                if(i->attached_files.size() > 0)
                {
                    Socket::Send("--VGhlIFNNVFAgUHJvdG9jb2wgaXMgY3JhcCBhbmQgd2FzIGludmVudGVkIGJ5IGFuIGlkaW90-- \r\n\r\n");
                }

                // end of mail
                Socket::Send(".\r\n");
            }
        }
        // disconnect
        Socket::Send("QUIT\r\n");
        Disconnect();
    }
    
    void SMTP::Send(EMail Mail)
    {
        Send(vector<EMail>(1,Mail));
    }

    string SMTP::ToBase64(const unsigned char* str, int length)
    {
        static string printable = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
        int fillbytes = (3-(length%3))%3;
        string result = "";
        char temp[5];
        temp[4]='\0';
        int i;
        for(i = 0; i <= length-3; i+=3)
        {
            temp[0] = printable[  str[i]/4                       ];
            temp[1] = printable[  (str[i]%4)*16 + str[i+1]/16    ];
            temp[2] = printable[  (str[i+1]%16)*4 + str[i+2]/64  ];
            temp[3] = printable[  str[i+2]%64                    ];
            result += string(&temp[0]);
        }

        switch(fillbytes)
        {
            case 0: break;
            case 1:
                temp[0] = printable[  str[i]/4                       ];
                temp[1] = printable[  (str[i]%4)*16 + str[i+1]/16    ];
                temp[2] = printable[  (str[i+1]%16)*4                ];
                temp[3] = '=';
                result += string(&temp[0]);
                break;
            case 2:
                temp[0] = printable[  str[i]/4                       ];
                temp[1] = printable[  (str[i]%4)*16                  ];
                temp[2] = '=';
                temp[3] = '=';
                result += string(&temp[0]);
                break;
        }

        return result;
    }

    string SMTP::ToBase64(string str)
    {
        return ToBase64((const unsigned char*)(str.c_str()), str.size());
    }

    string SMTP::basename(string str)
    {
        for(int i = str.size()-1; i >= 0; i--)
            if(str[i] == '/' || str[i] == '\\')
                return str.substr(i+1, str.size()-i-1);
        return str;
    }

}
